import { MessageTransmissionBase } from "./msg.transmission.base";
import { EventMessage, FisMessage, MessageTransmitter as MessageTransmitterInterface, TransmitterProfile } from '../interface/transport.interface'
import { AdapterSet, ConnectionAdaptorBase, ConnectionState, Event, Transport, TransportEvent, TransportMessage } from "../interface/connector.interface";
import { v4 as uuidv4 } from 'uuid'
import { TransmitterConnectionAdapter } from "../connector/connector.transmitter";
import { BehaviorSubject, distinct, filter, map, Observable, Subject } from "rxjs";
import { RetransmissionService } from "../utils/retransmission.service";
import { WrappedMessage } from "../utils/message.ordering";
import { ConnectionAdapter } from "../connector/connector.base";

/* Take in all the messages that needs to be transported, and divide them accordingly. So the connector instances will do just that
connectors or adapters will have their own identifier*/
export class MessageTransmissionTransmitter extends MessageTransmissionBase implements MessageTransmitterInterface {
    private messageToBeTransmitted!: Subject<FisMessage | WrappedMessage>
    transmitterProfile!: TransmitterProfile;
    retransmission!: RetransmissionService;

    constructor(profile: TransmitterProfile, adapter: TransmitterConnectionAdapter, event: Observable<TransportEvent>) {
        super(event)
        this.event = event
        this.messageToBeTransmitted = new Subject()
        this.retransmission = new RetransmissionService()
        this.setTransmitter(profile)
        this.setUpAdapter(adapter)
        this.setUpRetransmission()

        // special case just for http in case of server/client disconnected, the unsent msg will be flushed back into messageToBeTransmitted
        this.uniqueHandlerToFlushUnsentMessages(event)
    }

    // by the time this transmission set is instantiated, the connected client would've been online. Need ot manually signal retransmission to release buffer immediately
    setUpRetransmission(): void {
        let connectionStateEvent = new BehaviorSubject<'OFFLINE' | 'ONLINE'>('ONLINE')
        this.event.pipe(
            filter(event => event.event == 'New Client' || event.event == 'Client Disconnected' || event.event == 'Client Connected' || event.event == 'Server Disconnected' || event.event == 'Server Connected' || event.event == 'New Server'),
            filter(event => (event.data as EventMessage).clientId == this.transmitterProfile.id),
            map(event => {
                if (event.event == 'Client Disconnected' || event.event == 'Server Disconnected') {
                    return 'OFFLINE'
                } else {
                    return `ONLINE`
                }
            }),
            distinct()
        ).subscribe((signal: ConnectionState) => {
            connectionStateEvent.next(signal)
        })
        this.retransmission.implementRetransmission(this.messageToBeTransmitted, connectionStateEvent, true)
        // automatically subscribe to allow released bffered messages to be released
        this.retransmission.returnSubjectForBufferedItems().subscribe((message: WrappedMessage) => {
            // need to work with wrapped messages
            (this.mainAdapter as TransmitterConnectionAdapter).emit(message)
        })
    }

    setTransmitter(transmitterProfile: TransmitterProfile): void {
        this.transmitterProfile = transmitterProfile
    }

    emit(message: FisMessage): void {
        console.log(`Transmission Transmitter: `, message.header.messageID)
        this.messageToBeTransmitted.next(message)
    }

    setUpAdapter(adapter: ConnectionAdapter): void {
        // for now just hardcode to use 1 adapter type until connection manager is further enhacne to configure adapters on the fly
        this.mainAdapter = adapter
    }

    private uniqueHandlerToFlushUnsentMessages(event: Observable<TransportEvent>): void {
        event.pipe(
            filter(event => event.event == 'Re-Flush'),
            filter(event => (event.data as EventMessage).clientId == this.transmitterProfile.id),
        ).subscribe((event: TransportEvent) => {
            this.messageToBeTransmitted.next((((event.data as EventMessage).payload as TransportMessage).payload as WrappedMessage))
        })
    }

}